----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.all;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;
-- Entity
-----------------------------------------------------------


-- The PWM generator instantiates a timer that clocks at 40Mhz and outputs
-- when the internal counter reaches PG_I_timer (an N bit Generic)

-- At this point an PWM counter SA_cycle_count_current is incremented
-- If SA_cycle_count_current > PG_I_cycle_count_max then this counter is reset
-- If SA_cycle_count current < PG_I_duty_cycle then output 1
-- If SA_cycle_count current > PG_I_duty_cycle then output 0
-- The PWM counter can be reset by taking PG_I_cycle_synchhigh for one tick


entity pwm_generator is

	generic(
				PG_G_duty_cycle_bits: integer:= 8;
				PG_G_timer_bits: integer := 8
			  );
			  
	port	(
				PG_I_reset				 		: in	std_logic;
				PG_I_clk					 		: in	std_logic;
				PG_I_duty_cycle		 		: in	std_logic_vector (PG_G_duty_cycle_bits-1 downto 0);
				PG_I_cycle_count_max 		: in std_logic_vector (PG_G_duty_cycle_bits-1 downto 0);
				PG_I_timer				 		: in std_logic_vector (PG_G_timer_bits-1 downto 0);
				PG_I_cycle_synch				: in std_logic;
				PG_I_PWM_enable				: in std_logic;
				PG_I_full_On					: in std_logic;
				
				PG_I_Boost						: in unsigned(PG_G_duty_cycle_bits-1 downto 0);
				PG_I_BoostEnable				: in std_logic;
				
				PG_O_pwm_out			 		: out std_logic
			);
			
end pwm_generator;

architecture arch of pwm_generator is
	
	-- Signals
	
	signal PWM_TIME_NO_CONNECTION 	: std_logic_vector(PG_G_timer_bits-1 downto 0);
	signal SA_timer_complete				: std_logic;
	
	
	signal SA_duty_cycle						: unsigned(PG_G_duty_cycle_bits-1 downto 0);
	signal SA_db_duty_cycle					: unsigned(PG_G_duty_cycle_bits-1 downto 0);
	signal SA_db_duty_cycle_next			: unsigned(PG_G_duty_cycle_bits-1 downto 0);
	
	signal SA_cycle_count_max 				: unsigned(PG_G_duty_cycle_bits-1 downto 0);
	signal SA_db_cycle_count_max			: unsigned(PG_G_duty_cycle_bits-1 downto 0);
	signal SA_db_cycle_count_max_next	: unsigned(PG_G_duty_cycle_bits-1 downto 0);
	
	signal SA_cycle_count_next 			: unsigned(PG_G_duty_cycle_bits-1 downto 0);
	signal SA_cycle_count_current			: unsigned(PG_G_duty_cycle_bits-1 downto 0);

	signal SA_PWM_out_current				: std_logic;
	signal SA_PWM_out_next					: std_logic;
	signal SA_cycle_synch					: std_logic;

	type TA_state_type is 	(
									State_Initialise, State_Run
									);
	
	signal TA_state_current, TA_state_next: TA_state_type;

	begin	
	
	-- Instantiate a timer
	pwm_timer: entity work.mod_m_counter(arch)
	generic map (N=>PG_G_timer_bits)
	port map   (
					clk => PG_I_clk,
					reset => PG_I_reset,
					synch => SA_cycle_synch,
					timer_max => PG_I_timer,
					max_tick => SA_timer_complete,
					q => PWM_TIME_NO_CONNECTION
					);
					
					
	-- IO Mapping
	
	SA_cycle_count_max <= unsigned(PG_I_cycle_count_max);
	SA_cycle_synch <= PG_I_cycle_synch;
	
	
	PG_O_pwm_out <= (PG_I_PWM_enable AND (SA_PWM_out_current OR PG_I_full_on));

	process(PG_I_BoostEnable, PG_I_duty_cycle, PG_I_Boost)
	begin
		if (PG_I_BoostEnable = '0') then
			SA_duty_cycle <= unsigned(PG_I_duty_cycle);
		else
			SA_duty_cycle <= (unsigned(PG_I_duty_cycle) + PG_I_Boost);
		end if;
	end process;
	
	-- Clocking Process
	process(PG_I_reset, PG_I_clk)
	begin
		
		if(PG_I_reset = '1') then
			TA_State_Current <= State_Initialise;
			SA_cycle_count_current <= (others => '0');
			SA_pwm_out_current <= '0';
			
		elsif(PG_I_clk'event and PG_I_clk = '1') then
			SA_db_cycle_count_max <= SA_db_cycle_count_max_next;
			SA_db_duty_cycle <= SA_db_duty_cycle_next;
			SA_pwm_out_current <= SA_pwm_out_next;
			SA_cycle_count_current <= SA_cycle_count_next;
			TA_state_current <= TA_state_next;
			
		end if;
		
	end process;
	

	process(TA_State_Current, SA_duty_cycle, SA_cycle_count_max, SA_db_duty_cycle,
				SA_db_cycle_count_max, SA_timer_complete, SA_cycle_count_current,
				SA_cycle_synch,SA_PWM_out_current)
	
	begin
	
	SA_db_duty_cycle_next <= SA_db_duty_cycle;
	SA_db_cycle_count_max_next <= SA_db_cycle_count_max;
	SA_pwm_out_next <= SA_PWM_out_current;
	SA_cycle_count_next <= SA_cycle_count_current;
		
		case TA_state_current is
	
				when State_Initialise =>
					
					SA_db_duty_cycle_next <= SA_duty_cycle;
					SA_db_cycle_count_max_next <= SA_cycle_count_max;
					
					SA_cycle_count_next <= (others=> '0');
					
					if SA_duty_cycle = 0 then
						SA_pwm_out_next <= '0';
					else
						SA_pwm_out_next <= '1';
					end if;
					
					if SA_cycle_synch = '1' then
						TA_state_next <= State_Initialise;
					else
						TA_state_next <= State_Run;
					end if;
					
				when State_Run =>
				
					if SA_timer_complete  = '1' then
					
						if ((SA_cycle_count_current < SA_db_duty_cycle-1) and 
							(SA_db_duty_cycle /= 0)) then
								SA_PWM_out_next <= '1';
							else
								SA_PWM_out_next <= '0';
						end if;
						
						
						if SA_cycle_count_current < (SA_db_cycle_count_max -1) then
							SA_cycle_count_next <= SA_cycle_count_current + 1;
						else
							SA_cycle_count_next <= (others => '0');
								if SA_duty_cycle = 0 then
									SA_PWM_out_next <= '0';
								else
									SA_PWM_out_next <= '1';
								end if;
							SA_db_duty_cycle_next <= SA_duty_cycle;
							SA_db_cycle_count_max_next <= SA_cycle_count_max;
						end if;
												
					 end if;
					 
					
					if SA_cycle_synch = '1' then
						TA_state_next <= State_Initialise;
					else
						TA_state_next <= State_Run;
					end if;
					
							
		end case;
		
	end process;

end arch;
