-----------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
-----------------------------------------------------------


-- Entity
-----------------------------------------------------------

entity quadrature_counter is
	port(
		I_rst				: in	std_logic;
		I_clk100k		: in	std_logic;
		I_clk20m			: in	std_logic;
		I_rstToStart	: in  std_logic;
		--
		I_rst_position	: in	std_logic_vector(8 downto 0);
		I_index_offset	: in	std_logic_vector(8 downto 0);
		--I_reset_error	: in  std_logic;
		--
		I_ch_a			: in	std_logic;
		I_ch_b			: in	std_logic;
		I_index			: in	std_logic;
		I_IndexEnable	: in std_logic;
		I_IndexDirection	: in std_logic;
		--
		O_position		: out	std_logic_vector(8 downto 0);
		O_direction		: out	std_logic;
		
		O_DeGlitchCHA	: out std_logic;
		O_DeGlitchCHB	: out std_logic;
		O_DeGlitchIndex: out std_logic;
		--
		O_ErrorFlag		: out std_logic;
		O_error			: out	std_logic_vector(2 downto 0);
		O_error_value	: out	std_logic_vector(8 downto 0)
	);
end entity quadrature_counter;

-- Architecture
-----------------------------------------------------------
architecture arch of quadrature_counter is
	
	-- Components
	---------------------------
	
	component glitch_filter is
	
		generic	(
					G_bit_width : integer :=8
					);
					
		port		(
					I_rst			: in	std_logic;
					I_clk			: in	std_logic;
					--
					I_data_in	: in	std_logic;
					O_data_out	: out	std_logic
					);
					
	end component glitch_filter;
	
	-----------------------------
--	
	component edge_detector is
		port(
			I_rst			: in	std_logic;
			I_clk			: in	std_logic;
			--
			I_enable		: in	std_logic;
			I_edge_type	: in	std_logic;
			--
			I_data_in	: in	std_logic;
			O_edge_out	: out	std_logic
		);
	end component edge_detector;
--	
--	-----------------------------
--	
	component quadrature_decoder is
		generic(
			G_position_min	: integer := 0;
			G_position_max	: integer := 511
		);
		
		port(
			I_rst				: in	std_logic;
			I_clk				: in	std_logic;
			I_RstToStart	: in  std_logic;
			--
			I_rst_position	: in	std_logic_vector(8 downto 0);
			I_index_offset	: in	std_logic_Vector(8 downto 0);
			--I_reset_error	: in  std_logic;
			--
			I_ch_a			: in	std_logic;
			I_ch_b			: in	std_logic;
			I_index			: in	std_logic;
			--
			O_position		: out	std_logic_vector(8 downto 0);
			O_direction		: out	std_logic;
			--
			O_ErrorFlag		: out std_logic;
			O_error			: out	std_logic_vector(2 downto 0);
			O_error_value	: out	std_logic_vector(8 downto 0)
		);
	end component quadrature_decoder;
--	
--	-- Constants
--	-----------------------------
--	
--	constant CA_filter_bit_width	: positive range 2 to integer'high := 8;
--	
--	-- Signals
--	-----------------------------
--	
	signal SA_ch_a_dg			: std_logic;
	signal SA_ch_b_dg			: std_logic;
	signal SA_index_dg 		: std_logic;
--	
	signal SA_direction		: std_logic;
	signal SA_index_pulse	: std_logic;
	signal SA_output_error	: std_logic_vector( 8 downto 0);
--	
begin
--	
--	-- Instance port mapping
--	-----------------------------
	
	-- Channel A filter
	U_ch_a_filter: glitch_filter
		generic map(
						G_bit_width	=> 8
						)
		port map(
					I_rst			=> I_rst,
					I_clk			=> I_clk20m,
					--
					I_data_in	=> I_ch_a,
					O_data_out	=> SA_ch_a_dg
					);
	
--	-----------------------------
--	
--	-- Channel B filter
	U_ch_b_filter: glitch_filter
		generic map(
						G_bit_width	=> 8
						)
		port map	(
					I_rst			=> I_rst,
					I_clk			=> I_clk20m,
					--
					I_data_in	=> I_ch_b,
					O_data_out	=> SA_ch_b_dg
					);
	
--	-----------------------------
--	
	-- Index filter
	U_index_filter: glitch_filter
		generic map	(	
						G_bit_width	=> 8
						)
		port map(
					I_rst			=> I_rst,
					I_clk			=> I_clk20m,
					--
					I_data_in	=> I_index,
					O_data_out	=> SA_index_dg
					);
--	
--	-----------------------------
--	
	-- Index signal edge detector
	U_index_edge_detector: edge_detector
		port map(
			I_rst			=> I_rst,
			I_clk			=> I_clk100k,
			--
			I_enable		=> I_IndexEnable,
			I_edge_type	=> I_IndexDirection,
			--
			I_data_in	=> SA_index_dg,
			O_edge_out	=> SA_index_pulse
		);
--	
--	-----------------------------
--	
	-- Quadrature decoder
	U_quadrature_decoder: quadrature_decoder
		--generic(
		--			G_position_min	=> 0,
		--			G_position_max	=> 511
		--			);
		port map(
			I_rst				=> I_rst,
			I_clk				=> I_clk100k,
			I_RstToStart	=> I_rstToStart,
			--
			I_rst_position	=> I_rst_position,
			I_index_offset	=> I_index_offset,
			--I_reset_error	=> I_reset_error,	-- : in  std_logic;
			--
			I_ch_a			=> SA_ch_a_dg,
			I_ch_b			=> SA_ch_b_dg,
			I_index			=> SA_index_pulse,
			--
			O_position		=> O_position,
			O_direction		=> SA_direction,
			--
			O_ErrorFlag		=> O_ErrorFlag,
			O_error			=> O_error,
			O_error_value	=> SA_output_error
		);
--	
--	-- Concurrent signal
--	-- assignment
--	-----------------------------
--	
		O_direction <= SA_direction;
		O_error_value <= SA_output_error;
		O_DeGlitchIndex <= SA_index_dg;
		
		O_DeGlitchCHA	<=	SA_ch_a_dg;
		O_DeGlitchCHB	<= SA_ch_b_dg;
	
end arch;
