----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    11:51:50 17/10/2010 
-- Design Name: 
-- Module Name:    spi_rx_005 - arch
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

-- This version is more generic. It will have a defined data buffer size
-- and the sclk data edge will be settable

-- **********************************************************************
-- ** SPI_DataWidth must be < 128 bits otherwise the module will not work
-- Increase the width u_spi_counter_now & next if more than 128 bits 
-- need to be transmitted.
-- **********************************************************************

-- **********************************************************************
-- Version rx_005 -- 21 Oct 2010

-- This version has two more inputs added:

-- i_spi_tx_gate_sclk
-- If gate_sclk = 1 then sclk stops running when o_spi_rx_synch is high

-- i_spi_tx_gate_value
-- This is the value that sclk takes when o_spi_rx_synch is high 
-- if sclk is gated.
-- **********************************************************************

-- **********************************************************************
-- Version rx_004 -- 17 Oct 2010

-- The requirement that start_read must be presented when
-- i_spi_rx_en = '1' has been removed so that the circuit
-- can interface with other modules not throttled back to 
-- fpga_clk/2n  
-- **********************************************************************

-- **********************************************************************
-- Version rx_003 -- 22 Sept 2010

-- This version has an enable input that allows the clock rate of the 
-- spi interface to be throttled back to fpga_clk/2n where 
-- n is controlled via a mod m counter if n>1.

-- Note that the start_read signal must be presented when 
-- i_spi_rx_en = '1' else it will be ignored.
-- **********************************************************************

entity spi_rx_005 is

	generic	(
				g_SPI_DataWidth		: natural:= 8  -- number of bits 
				);

    Port (
				clk 						: in std_logic;
				reset 					: in std_logic;
				i_spi_rx_sclk_edge	: in std_logic;
				
				-- 0 = din is sampled on the Rising Edge of sclk
				--		(din changes on the Falling Edge of sclk?)
				-- ***
				-- 1 = din is sampled on the Falling Edge of sclk
				-- 	(din changes on the Rising Edge of sclk?)
				
				-- For the AD7685 din Changes on the Falling Edge of sclk
				
				i_spi_rx_start_read	: in std_logic;
				i_spi_rx_din			: in std_logic;
				
				i_spi_rx_en				: in std_logic;
				
				i_spi_rx_gate_sclk		: in std_logic;
				-- 1 = sclk stopped when synch is high
				-- 0 = sclk runs continously
				
				i_spi_rx_gate_value		: in std_logic;
				-- 1 = sclk defaults to '1' when gated
				-- 0 = sclk defaults to '0' when gated
				
				o_spi_rx_busy			: out std_logic;
				o_spi_rx_data			: out std_logic_vector(g_SPI_DataWidth - 1 downto 0);		
			
				o_spi_rx_sclk			: out std_logic;
				o_spi_rx_synch			: out std_logic	
			);
			
end spi_rx_005;


architecture arch of spi_rx_005 is

	type T_state_type is 	(
										State_Wait,
										State_Start,
										State_Count								
									);
							
	signal T_state_now, T_state_next: T_state_type;
	
	
	constant DELAY : natural := (2*g_SPI_DataWidth);


	signal sl_d_in						: std_logic;
	signal sl_q_out					: std_logic;
	signal sl_qbar_out				: std_logic;
	signal sl_start_read				: std_logic;
	signal sl_busy						: std_logic;
	
	signal sl_start_read_int_now	: std_logic;
	signal sl_start_read_int_next : std_logic;

	signal slv_shift_reg_now		: std_logic_vector(g_SPI_DataWidth - 1 downto 0);
	signal slv_shift_reg_next 		: std_logic_vector(g_SPI_DataWidth - 1 downto 0);
	signal slv_data_buffer_now		: std_logic_vector(g_SPI_DataWidth - 1 downto 0);
	signal slv_data_buffer_next	: std_logic_vector(g_SPI_DataWidth - 1 downto 0);
	
	signal u_spi_counter_now		: unsigned (7 downto 0);
	signal u_spi_counter_next 		: unsigned (7 downto 0);
	
	signal sl_sclk_out_int			: std_logic;
	signal sl_spi_rx_synch_int		: std_logic;

begin

	sl_d_in <= not(sl_q_out);
	sl_start_read <= i_spi_rx_start_read;
	o_spi_rx_busy <= sl_busy;
	o_spi_rx_data <= slv_data_buffer_now;

	o_spi_rx_synch <= sl_spi_rx_synch_int;
	
	-- Select whether to output sclk directly or its inverse
	-- This is determined by the value of the input 
	-- i_spi_rx_sclk_edge
	process (i_spi_rx_sclk_edge, sl_q_out, sl_qbar_out)
	begin
		if i_spi_rx_sclk_edge = '1' then
			sl_sclk_out_int <= sl_q_out;
		else
			sl_sclk_out_int <= sl_qbar_out;
		end if;
	end process;
	
	
	--Select whether output is gated
	process(sl_spi_rx_synch_int, i_spi_rx_gate_value, sl_sclk_out_int, i_spi_rx_gate_sclk)
	begin
		
		o_spi_rx_sclk <= sl_sclk_out_int;
	
		if (sl_spi_rx_synch_int = '1' and i_spi_rx_gate_sclk = '1') then
			o_spi_rx_sclk <= i_spi_rx_gate_value;
		end if;
	end process;

	process(clk, reset)
			
	begin
			
		if (reset = '1') then
			sl_q_out <= '0';
			sl_qbar_out <= '1';
			slv_shift_reg_now<= (others => '0');
			u_spi_counter_now <= (others => '0');
			slv_data_buffer_now <= (others => '0');
			T_state_now <= State_Wait;
			sl_start_read_int_now <= '0';
		elsif (clk'event and clk = '1') then
			sl_start_read_int_now <= sl_start_read_int_next;
			if (i_spi_rx_en = '1') then
				sl_q_out <= sl_d_in;
				sl_qbar_out <= sl_q_out;
				T_state_now <= T_state_next;
				u_spi_counter_now <= u_spi_counter_next;	
				slv_shift_reg_now<= slv_shift_reg_next;
				slv_data_buffer_now <= slv_data_buffer_next;
			end if;
		end if;
				
	end process;
	
	-- Synchronise external i_spi_rx_start_read with rx state 
	-- machine.
	-- This is necessary as the state machine clk is effectively 
	-- and'ed with the enable input
	process(sl_start_read, sl_start_read_int_now, T_state_now)
	begin
		sl_start_read_int_next <= sl_start_read_int_now;
	
		if sl_start_read = '1' and T_state_now = State_Wait then
			sl_start_read_int_next <= '1';
		elsif T_state_now /= State_Wait then
			sl_start_read_int_next <= '0';
		end if;
		
	end process;


	-- Use a small state Machine to orchestrate the whole process
	process(T_state_now, T_state_next, slv_shift_reg_now, u_spi_counter_now, 
		sl_start_read_int_now, sl_q_out, i_spi_rx_din, sl_start_read, slv_data_buffer_now)
	
	begin
	
		T_state_next <= T_state_now;
		sl_spi_rx_synch_int <= '1';
		slv_shift_reg_next <= slv_shift_reg_now;
		slv_data_buffer_next <= slv_data_buffer_now;
		sl_busy <= '1';
		
		if u_spi_counter_now > 0 then
			u_spi_counter_next <= (u_spi_counter_now - 1);
		else
			u_spi_counter_next <= to_unsigned(0,8);
		end if;
		
		case T_state_now is
		
			when State_Wait =>
			
				T_state_next <= State_Wait;
				slv_shift_reg_next <= std_logic_vector(to_unsigned(0,g_SPI_DataWidth));
			
				if sl_start_read_int_now = '0' then
					sl_busy <= '0';
				else		
					T_state_next <= State_Start;
				End if;
					
			when State_Start =>
									
				-- Ensure Synchronisation with SCLK
				
				-- synch signal is set to 0 on the rising edge of the internal sclk, 
				-- reading of din is done on the falling edge of the internal sclk.
				
				-- Remember that the setting of i_spi_rx_sclk_edge determines whether
				-- the SPI tx is actually seeing a rising or a falling edge
				
				If sl_q_out = '0' then -- internal sclk
					u_spi_counter_next <= to_unsigned(DELAY,8);
					slv_shift_reg_next <= std_logic_vector(to_unsigned(0,g_SPI_DataWidth));
					slv_data_buffer_next <= std_logic_vector(to_unsigned(0,g_SPI_DataWidth));
					T_state_next <= State_Count;
				else
					T_state_next <= State_Start;
				End If;
			
			
			when State_Count =>
				
				sl_spi_rx_synch_int <= '0';	
				
				if (u_spi_counter_now > 0) then
					if (sl_q_out) = '1' then
						slv_shift_reg_next <= slv_shift_reg_now(g_SPI_DataWidth-2 downto 0) & i_spi_rx_din;
					end if;
					T_state_next <= State_Count;
				else
					--Synchronise to sclk & write to fifo
					if sl_q_out = '1' Then
						T_state_next <= State_Wait;
						slv_data_buffer_next <= slv_shift_reg_now;
						-- Override the SPI_O_synch setting
						sl_spi_rx_synch_int <= '1';	
					else
						T_state_next <= State_Count;
					end if;
				end if;
				
		end case;
		
	end process;

end arch;

