----------------------------------------------------------------------------------
-- Company: 		Paul J Smith
-- Engineer: 
-- 
-- Create Date:    11:51:50 08/06/2010 
-- Design Name: 
-- Module Name:    spi_tx_004 - arch
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;
-- **********************************************************************
-- version tx_004 -- 21 Oct 2010

-- This version has two more inputs added:

-- i_spi_tx_gate_sclk
-- If gate_sclk = 1 then sclk stops running when o_spi_tx_synch is high

-- i_spi_tx_gate_value
-- This is the value that sclk takes when o_spi_tx_synch is high 
-- if sclk is gated.

-- **********************************************************************
-- version tx_003 -- 22 Sept 2010

-- This version has an enable input that allows the clock rate of the 
-- spi interface to be throttled back to fpga_clk/2n where 
-- n is controlled via a mod m counter if n>1.
-- **********************************************************************
-- Version tx_004

entity spi_tx_004 is

	generic	(
				g_SPI_Data_Width		: natural:= 8  -- number of bits 
				);

    Port (
				clk		 					: in std_logic;
				reset			 				: in std_logic;
				
				i_spi_tx_fifo_data_in	: in std_logic_vector(g_SPI_Data_Width-1 downto 0);
				i_spi_tx_fifo_write		: in std_logic;
				i_spi_tx_sclk_edge		: in std_logic;
				-- 1 = Change tx data on Rising Edge of sclk
				-- 0 = Change tx data on Falling Edge of sclk
				
				i_spi_tx_en					: in std_logic;
				
				i_spi_tx_gate_sclk		: in std_logic;
				-- 1 = sclk stopped when synch is high
				-- 0 = sclk runs continously
				
				i_spi_tx_gate_value		: in std_logic;
				-- 1 = sclk defaults to '1' when gated
				-- 0 = sclk defaults to '0' when gated
	
				o_spi_tx_fifo_full		: out std_logic;	

				-- SPI outputs
				o_spi_tx_dout				: out std_logic;
				o_spi_tx_sclk				: out std_logic;
				o_spi_tx_synch				: out std_logic	
			);
			
end spi_tx_004;


architecture arch of spi_tx_004 is

	type T_state_type is 	(
										State_Wait,
										State_Start,
										State_Count								
									);
							
	signal T_state_now, T_state_next: T_state_type;

	signal sl_d_in					: std_logic;
	signal sl_q_out				: std_logic;
	signal sl_qbar_out			: std_logic;
	signal sl_fifo_empty 		: std_logic;
	signal sl_fifo_read			: std_logic;
	
	signal slv_write_data		: std_logic_vector(g_SPI_Data_Width-1 downto 0);
	signal slv_read_data			: std_logic_vector(g_SPI_Data_Width-1 downto 0);
	signal slv_shift_reg_now	: std_logic_vector(g_SPI_Data_Width-1 downto 0);
	signal slv_shift_reg_next 	: std_logic_vector(g_SPI_Data_Width-1 downto 0);
	
	signal sl_sclk_out_int		: std_logic;
	signal sl_spi_tx_synch_int	: std_logic;
	
	signal u_spi_counter_now	: unsigned (7 downto 0);
	signal u_spi_counter_next : unsigned (7 downto 0);
	
begin

	fifo_dataBuffer: entity work.fifo_fwft_001(arch)
		generic map (Data_Bits => g_SPI_Data_Width, Address_Bits => 2)
		port map(
			clk				=> clk,
			reset				=> reset,
			i_fifo_rd		=> sl_fifo_read,
			i_fifo_wr		=> i_spi_tx_fifo_write,
			i_fifo_wr_data	=> slv_write_data,
			o_fifo_empty	=> sl_fifo_empty,
			o_fifo_full		=> o_spi_tx_fifo_full,
			o_fifo_rd_data	=>	slv_read_data
	);

	-- Assignments
	sl_d_in <= not(sl_q_out);
	sl_qbar_out <= not(sl_q_out);
	slv_write_data <= i_spi_tx_fifo_data_in;
	
	o_spi_tx_dout <= slv_shift_reg_now(g_SPI_Data_Width-1);
	o_spi_tx_synch <= sl_spi_tx_synch_int;

	process(clk, reset)
			
	begin
			
		if (reset = '1') then
			sl_q_out <= '0';
			slv_shift_reg_now <= (others => '0');
			u_spi_counter_now <= (others => '0');
			T_state_now <= State_Wait;
		elsif (clk'event and clk = '1') then
			if (i_spi_tx_en = '1') then						
				sl_q_out <= sl_d_in;
				T_state_now <= T_state_next;
				u_spi_counter_now <= u_spi_counter_next;	
				slv_shift_reg_now <= slv_shift_reg_next;		
			end if;
		end if;
				
	end process;
	
	--Select whether output changes on +ve or -ve sclk
	process (i_spi_tx_sclk_edge, sl_q_out, sl_qbar_out)
	begin
		if i_spi_tx_sclk_edge = '1' then
			sl_sclk_out_int <= sl_q_out;	
		else
			sl_sclk_out_int <= sl_qbar_out;
		end if;
	end process;

	--Select whether output is gated
	process(sl_spi_tx_synch_int, i_spi_tx_gate_value, sl_sclk_out_int, i_spi_tx_gate_sclk)
	begin
		
		o_spi_tx_sclk <= sl_sclk_out_int;
	
		if (sl_spi_tx_synch_int = '1' and i_spi_tx_gate_sclk = '1') then
			o_spi_tx_sclk <= i_spi_tx_gate_value;
		end if;
	end process;
	
	-- A state Machine to orchestrate the whole process
	process(T_state_now, T_state_next, sl_fifo_empty, slv_shift_reg_now, 
		slv_read_data, u_spi_counter_now, sl_q_out, i_spi_tx_en)
	
	begin
	
		T_state_next <= T_state_now;
		sl_fifo_read <= '0';
		sl_spi_tx_synch_int <= '1';
		slv_shift_reg_next <= slv_shift_reg_now;
		
		
		if u_spi_counter_now > 0 then
			u_spi_counter_next <= (u_spi_counter_now - 1);
		else
			u_spi_counter_next <= X"00";
		end if;
		
		case T_state_now is
		
			when State_Wait =>
				
				T_state_next <= State_Wait;
				
					if (sl_fifo_empty = '0') Then			
						T_state_next <= State_Start;
					end if;
				
			when State_Start =>
									
				-- Ensure Synchronisation with SCLK
				if (sl_q_out = '0' and i_spi_tx_en = '1') then
					u_spi_counter_next <= to_unsigned(2*g_SPI_Data_Width,8);
					T_state_next <= State_Count;
					slv_shift_reg_next <= slv_read_data;
					-- Read signals must be one clock pulse long only
					-- else multiple reads will occur
					sl_fifo_read <= '1';
				else
					T_state_next <= State_Start;
				end if;
			
			
			when State_Count =>
				
				sl_spi_tx_synch_int <= '0';	
				
				if (u_spi_counter_now > 0) then
					if (sl_q_out = '0') then
						slv_shift_reg_next <= slv_shift_reg_now(g_SPI_Data_Width-2 downto 0) & '0';
					end if;
					T_state_next <= State_Count;
				else
					--Synchronise to sclk
					if (sl_q_out = '1') then
						T_state_next <= State_Wait;
						sl_spi_tx_synch_int <= '1';
					else
						T_state_next <= State_Count;
					end if;
				end if;
				
		end case;
		
	end process;

end arch;

